class Globals:
    
    DEBUG = False

    # The Parent of all widgets object stored here
    # from `runApp`'s first parameter.
    __PARENT__ = None

    # The current pathname location used for routing.
    __LOC__ = None

    # Dictionary of mapped routes: {"/route": Component()}
    __ROUTES__ = []

    # All the event objects for firing.
    __EVENTS__ = {}

__all__ = ['Events', 'Globals']


#!/usr/bin/env python3

import os, sys, random, string, time
from shutil import copytree, rmtree, copy
from distutils.dir_util import copy_tree

from ..shared.pythonMinifier import minify

def execute_from_command_line(argv=None):
    utility = ManagementUtility()

    """Entry Point"""
    try:
        if sys.argv[1] == "create-app":
            try:
                name = sys.argv[2]
            except IndexError:
                name = "MyApp"
            
            try:
                description = sys.argv[3]
            except IndexError:
                description = "PyFyre web application."
            
            utility.create_app(name, description)
        elif sys.argv[1] == "runapp":
            try:
                port = sys.argv[2]
            except IndexError:
                port = None

            try:
                directory = sys.argv[3]
            except IndexError:
                directory = None

            utility.run_app(directory, port)
        elif sys.argv[1] == "build":
            try:
                directory = sys.argv[2]
            except IndexError:
                directory = None

            utility.build_app(directory)
        elif sys.argv[1] == "help":
            utility.pyfyre_help()
        else:
            utility.pyfyre_help()
    except IndexError as e:
        utility.pyfyre_help()

class ManagementUtility:
    def pyfyre_help(self):
        PYFYRE_HELP = """Manage your PyFyre projects.

        Common Commands:

            pyfyre.py create-app [name] [description]
                Create a new PyFyre project in your current directory.

            pyfyre.py runserver [port=8080]
                Run a live server in your current directory.

            pyfyre.py help
                Show this message.
        """
        print(PYFYRE_HELP)

    def create_app(self, app_name: str, app_description: str):
        """Create a PyFyre project."""
        
        print("Creating your PyFyre project...")
        
        path = os.path.join(os.getcwd(), app_name)
        
        try:
            os.makedirs(path)
        except FileExistsError:
            prompt = input(f"Project already exists. Want to overwrite the project '{app_name}'? (y or n): ")

            if prompt == "y":
                rmtree(path)
                os.makedirs(path)
            else:
                print("Aborting...")
                return
        
        # copy the `core` directory contents to the user's project directory
        core_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
        copy_tree(core_dir, os.path.join(path, "pyfyre"))
        
        # copy the `user` directory contents to the user's project directory
        user_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "user"))
        copy_tree(user_dir, path)

        os.mkdir(os.path.join(path, "public"))

        copy(os.path.join(path, "favicon.ico"), os.path.join(path, "public"))
        os.remove(os.path.join(path, "favicon.ico"))
        
        # edit the `index.html` file
        with open(os.path.join(user_dir, "index.html")) as file:
            content = file.read().format(app_name=app_name, app_description=app_description, main_key="{main_key}")
        with open(os.path.join(path, "public", "index.html"), "w") as file:
            file.write(content)
        
        # edit the `README.md` file
        with open(os.path.join(user_dir, "README.md")) as file:
            content = file.read().format(app_name=app_name, app_description=app_description)
        with open(os.path.join(path, "README.md"), "w") as file:
            file.write(content)
        
        # edit the `settings.py` file
        with open(os.path.join(user_dir, "settings.py")) as file:
            content = file.read().format(app_name=app_name, app_description=app_description)
        with open(os.path.join(path, "settings.py"), "w") as file:
            file.write(content)

        # edit the `.gitignore` file
        with open(os.path.join(path, ".gitignore"), "w") as file:
            file.write(__GITIGNORE__)

        os.chdir(os.path.join(path, "pyfyre"))

        os.system("brython-cli --install")

        # Remove unused files produced by Brython
        os.remove("demo.html")
        os.remove("unicode.txt")
        os.remove("README.txt")
        os.remove("brython.js")
        os.remove("index.html")

        rmtree(os.path.join(path, "pyfyre", "management"))
        rmtree(os.path.join(path, "pyfyre", "user"))
        rmtree(os.path.join(path, "pyfyre", "shared"))

        self.minify_dir(os.path.join(path, "pyfyre"))

        os.system("brython-cli --make_package pyfyre")
        os.remove("pyfyre.py")
        os.system("brython-cli --modules")

        os.chdir(os.path.join(".."))
        os.mkdir("pyf_modules")

        with open(os.path.join(path, "pyfyre", "brython_modules.js")) as file:
            content = file.read()
        with open(os.path.join("pyf_modules", "builtins.js"), "w") as file:
            file.write(content)
        with open(os.path.join(path, "pyfyre", "pyfyre.brython.js")) as file:
            content = file.read()
        with open(os.path.join("pyf_modules", "modules.js"), "w") as file:
            file.write(content)
        
        rmtree("pyfyre")
        os.remove("__init__.py")
        os.remove("index.html")
        os.remove(os.path.join("styles", "__init__.py"))

        try: rmtree("__pycache__")
        except: ...

        try: rmtree(os.path.join("src", "__pycache__"))
        except: ...

        try: rmtree(os.path.join("styles", "__pycache__"))
        except: ...

        os.system("cls" if os.name == "nt" else "clear")
        
        print("Project created successfully.")

    def produce(self, directory_path, build_path, reload=False):

        # If it's in the entry point, create a new path for the first time and produce
        # the essential Python js module. This will make the Hot Reload more faster.
        if not reload:
            copytree(directory_path, build_path)

            for _, __, filenames in os.walk(os.path.join(build_path, "public")):
                for filename in filenames:
                    copy(os.path.join(build_path, "public", filename), build_path)

            for _, __, filenames in os.walk(os.path.join(build_path, "pyf_modules")):
                for filename in filenames:

                    _, ext = os.path.splitext(filename)

                    if ext == ".js":
                        with open(os.path.join(directory_path, "pyf_modules", filename)) as file:
                            content = file.read()
                        with open(os.path.join(build_path, filename), "w") as file:
                            file.write(content)

            # Refresh the css files
            try:
                css_path = os.path.join(directory_path, "styles")
                src_css_path = os.path.join(build_path, "styles")
                
                copytree(css_path, src_css_path)
            except:
                print("Warning: `styles` folder cannot be found.")

        with open(os.path.join(build_path, "public", "index.html") if not reload else os.path.join(build_path, "index.html")) as file:
            index_content = file.read()

        # Produce the necessary standard library for Brython
        os.chdir(os.path.join(directory_path, "src"))
        os.system("brython-cli --make_package src")

        # Write files to build directory
        with open(os.path.join("src.brython.js")) as file:
            content = file.read()
        with open(os.path.join(build_path, "src.brython.js"), "w") as file:
            file.write(content)
        with open(os.path.join(build_path, "index.html"), "w") as file:
            file.write(index_content)
        with open(os.path.join(directory_path, "src", "__init__.py")) as file:
            content = file.read()
        with open(os.path.join(build_path, "__init__.py"), "w") as file:
            file.flush()
            file.write(content)

        os.remove("src.brython.js")

        os.chdir(build_path)
        
        # Remove the reloading files for refresh
        if reload:

            _ignores = set(["serve", "__dev__"])

            # Remove the src files
            for _, dirs, filenames in os.walk(build_path):
                [dirs.remove(tmp) for tmp in list(dirs) if tmp in _ignores]

                for filename in filenames:
                    name, ext = os.path.splitext(filename)

                    if ext == ".js":
                        if "main" in name:
                            with open(os.path.join(build_path, "src.brython.js")) as file:
                                content = file.read()
                            with open(filename, "w") as file:
                                file.write(content)

            try:
                # Refresh the css files
                css_path = os.path.join(directory_path, "styles")
                src_css_path = os.path.join(build_path, "styles")

                rmtree(src_css_path)

                src_css_path = os.path.join(build_path, "styles")

                copytree(css_path, src_css_path)
            except FileNotFoundError:
                print("Warning: `styles` folder cannot be found.")

        # Remove unnecessary files
        if not reload:
            try: rmtree("public")
            except Exception: ...

            try: rmtree("__pycache__")
            except Exception: ...

            try: rmtree(os.path.join("src", "__pycache__"))
            except Exception: ...

            try: rmtree(os.path.join("styles", "__pycache__"))
            except Exception: ...
            
            try: os.remove("requirements.txt")
            except Exception: ...

            try: os.remove("settings.py")
            except Exception: ...
            
            try: os.remove("runtime.txt")
            except Exception: ...

            try: os.remove(".gitignore")
            except Exception: ...

            rmtree("pyf_modules")
            os.remove("README.md")

            rmtree("src")

    def run_app(self, directory, port):
        print("Running your app in a development server...")

        try:
            from livereload import Server
        except ImportError:
            raise ImportError("Cannot find the liveserver module. Is it installed?")

        server = Server()

        _directory = os.path.abspath(directory) if directory else os.getcwd()
        
        if os.path.exists(os.path.join(_directory, "_pyfyre")):
            rmtree(os.path.join(_directory, "_pyfyre"))
        
        _build = os.path.join(_directory, "_pyfyre")

        def reload():
            print("Detected file changes, performing hot reload...")
            self.produce(_directory, _build, reload=True)
            self.produceVfsBundle(_build, _directory, reload=True)
            print("Hot reload successful!")

        self.produce(_directory, _build)
        self.produceVfsBundle(_build)

        os.system("cls" if os.name == "nt" else "clear")
        
        print("Happy Hacking!")

        server.watch(f"{_directory}/src/", reload)
        server.watch(f"{_directory}/styles/", reload)
        server.serve(port=port if port else 8000, host="localhost", root=os.path.join(_directory, "_pyfyre"))

    def build_app(self, directory):
        print("Producing optimized build for your project...")

        directory_path = os.path.abspath(directory) if directory else os.getcwd()

        if os.path.exists(os.path.join(directory_path, "build")):
            rmtree(os.path.join(directory_path, "build"))

        build_path = os.path.join(directory_path, "build")

        self.produce(directory_path, build_path)
        self.produceVfsBundle(build_path)

        print("Build succeeded!")

    def produceVfsBundle(self, build_path, directory_path=None, reload=False):

        try:
            import js2py
        except ImportError:
            raise ImportError("Cannot find js2py. Is it installed?")

        os.chdir(build_path)

        vfs = {}

        ctx_main = js2py.EvalJs()
        ctx_std = js2py.EvalJs()
        ctx_pyf = js2py.EvalJs()
        
        with open(os.path.join(build_path, "src.brython.js")) as file:
            content = file.readlines()
            content.pop(0)
            content.pop(1)
            main_js = ''.join(content)
        with open(os.path.join(build_path, "builtins.js") if not reload else os.path.join(directory_path, "pyf_modules", "builtins.js")) as file:
            content = file.readlines()
            content.pop(0)
            content.pop(0)
            content.pop(1)
            std_js = ''.join(content)
        with open(os.path.join(build_path, "modules.js") if not reload else os.path.join(directory_path, "pyf_modules", "modules.js")) as file:
            content = file.readlines()
            content.pop(0)
            content.pop(1)
            pyf_js = ''.join(content)
        with open(os.path.join("..", "settings.py")) as file:
            settings = file.read()

        ctx_main.execute(main_js)
        ctx_std.execute(std_js)
        ctx_pyf.execute(pyf_js)

        main_scripts = ctx_main.scripts.to_dict()
        std_scripts = ctx_std.scripts.to_dict()
        pyf_scripts = ctx_pyf.scripts.to_dict()

        def appendVfs(sc, py_script=False, name=None, requires=None):
            if py_script:
                vfs[f"{name}"] = [".py", sc, requires]
                return

            for k, v in list(sc.items()):
                if k == "$timestamp":
                    sc.pop(k)
                
                vfs[k] = v

        appendVfs(main_scripts)
        appendVfs(std_scripts)
        appendVfs(pyf_scripts)
        appendVfs(settings, py_script=True, name="pyfyre.settings", requires=["pyfyre.globals"])

        main_key = ''.join(random.choice(string.ascii_lowercase) for i in range(15))

        if reload:
            for _, _, files in os.walk(build_path):
                for file in files:
                    name, ext = os.path.splitext(file)

                    if ext == ".js":
                        if "main" in name:
                            name = name.split('.')
                            main_key = name[-1]

        with open(os.path.join(build_path, f"main.{main_key}.js"), "w", encoding="utf-8") as file:
            vfs["$timestamp"] = int(1000 * time.time())

            brython = [
                "__BRYTHON__.use_VFS = true;\n",
                "var PyFyreDOM = {isDOMLoaded: false,listeners: [],addListener: (fun) => {PyFyreDOM.listeners.push(fun);},broadcast: () => {for (var i = 0; i < PyFyreDOM.listeners.length; i++) {PyFyreDOM.listeners[i]()}}}\n",
                f"var scripts = {str(vfs)}",
                "\n__BRYTHON__.update_VFS(scripts)"
            ]
            
            file.writelines(brython)

        with open(os.path.join(build_path, "index.html")) as file:
            content = file.read().format(main_key=main_key)
        with open(os.path.join(build_path, "index.html"), "w") as file:
            file.write(content)

        with open(os.path.join(build_path, "__init__.py")) as file:
            minified = minify(file.read())
        with open(os.path.join(build_path, "__init__.py"), "w") as file:
            file.write(minified)

        os.remove(f"src.brython.js")

        if not reload:
            os.remove(f"modules.js")
            os.remove(f"builtins.js")

    def minify_dir(self, path):
        for dirpath, _, filenames in os.walk(path):
            for filename in filenames:
                _, ext = os.path.splitext(filename)

                if ext == ".py":
                    with open(os.path.join(dirpath, filename)) as file:
                        minified = minify(file.read())
                    with open(os.path.join(dirpath, filename), "w") as file:
                        file.write(minified)

__GITIGNORE__ = """# Byte-compiled / optimized / DLL files
__pycache__/
*.py[cod]
*$py.class

# PyFyre
build/
_pyfyre/

# Environments
.env
.venv
env/
venv/
ENV/
env.bak/
venv.bak/"""

if __name__ == "__main__":
    execute_from_command_line()



from pyfyre.widgets import *

class Header(Container):
    def __init__(self):

        _nav_items = [
            ["Documentation", "https://pyfyre.gitbook.io/docs/"],
            ["Playground", "/playground"],
            # ["Cookbook", "/"]
        ]

        def nav_items(i):
            return Link(
                className="block py-2 pr-4 pl-3 text-base text-black md:bg-transparent md:text-black md:p-0 cursor-pointer",
                textContent=_nav_items[i][0],
                to=_nav_items[i][1],
                external=True if "https" in _nav_items[i][1] else False
            )

        super().__init__(
            className = "bg-white border-gray-200 px-2 sm:px-4 py-2.5 border-b-2 sticky top-0",
            children = [
                Container(
                    className="flex flex-wrap justify-between items-center mx-auto",
                    children=[
                        Container(
                            className="flex",
                            children=[
                                Image(
                                    className="w-10",
                                    src="https://scontent.fmnl13-1.fna.fbcdn.net/v/t1.6435-9/164540820_110370211141719_22088476788963606_n.jpg?_nc_cat=104&ccb=1-5&_nc_sid=09cbfe&_nc_ohc=Qg8QkUIRCLIAX9ll7Gs&tn=LaPlocOuMCICDrK8&_nc_ht=scontent.fmnl13-1.fna&oh=00_AT8iHOF8SKnb4kNw1oUH24roDJ3w7HQdux72Jdf2850yrg&oe=62169F42"
                                ),
                            ]
                        ),
                        Container(
                            className="flex items-center space-x-5",
                            children=[
                                ListBuilder(
                                    className="invisible flex space-x-5 lg:visible",
                                    count=len(_nav_items),
                                    builder=nav_items
                                ),
                                Link(
                                    className="text-white cursor-pointer font-black bg-[#fab327] focus:ring-4 font-medium rounded-3xl text-sm px-5 py-2.5 text-center mr-3 md:mr-0",
                                    textContent="Get Started",
                                    to="https://pyfyre.gitbook.io/docs/quick-start",
                                    external=True
                                )
                            ]
                        ),
                    ]
                )
            ]
        )


from pyfyre.core.exceptions import InvalidController
from pyfyre.widgets.widget import Widget

class TextInput(Widget):
    """Creates a Text input widget.

    Parameters
    ----------
    controller : TextInputController
        TextInputController method allows you to leverage `input` HTML built-ins.

    onInput : method
        Method passed on this parameter will be called automatically
        when the user puts input to this widget.

    defaultValue : str
        Default value of the TextInput
    """
    
    def __init__(self, controller=None, onInput=None, defaultValue="", className="", multiline=False, props: dict=None):
        super().__init__("input" if not multiline else "textarea", className=className, props=props)
        self.controller = controller
        self.onInput = onInput
        self.defaultValue = defaultValue
    
    def dom(self):
        element = super().dom()

        element.value = self.defaultValue
        
        if self.controller:
            self.controller.callback(self)

            def bindController(event):
                self.controller.state = element.value

            element.bind('input', bindController)

            if self.controller.state:
                element.value = self.controller.state
        
        if self.onInput:
            
            def wrapper(event):
                self.onInput(element.value)
                
            element.bind('input', wrapper)
        
        return element

class TextInputController:
    """
    TextInputController, used to leverage HTML `input` built-ins.
    
    To use this, you must initialize this on the `__init__` method of
    your component and pass the variable on the TextInput as a controller.

    Methods
    -------
    changeAttribute(readOnly=False, disabled=False)
        Change an attribute of TextInput.

    Properties
    ----------
    value -> str
        returns a string of the current value of the TextInput
    """

    def __init__(self):
        self.this = None
        self.state = None
        
        # TextInput attributes
        self.readOnly = False
        self.disabled = False

    def changeAttribute(self, readOnly=False, disabled=False):
        if not self.this or not isinstance(self.this, TextInput):
            raise InvalidController("Looks like you haven't called this controller as a parameter controller of TextInput or you provided an invalid controller.")

        self.readOnly = readOnly
        self.disabled = disabled

        if self.readOnly:
            self.this.element.attrs['readonly'] = None
        else:
            if 'readonly' in self.this.element.attrs:
                del self.this.element.attrs['readonly']

        if self.disabled:
            self.this.element.attrs['disabled'] = None
        else:
            if 'disabled' in self.this.element.attrs:
                del self.this.element.attrs['disabled']

    def callback(self, this: TextInput):
        """
        This is called when this TextInputController object
        is passed as a parameter Controller of TextInput widget.
        """
        self.this = this
        
    @property
    def value(self):
        return self.this.element.value

    def setValue(self, newValue):
        self.state = newValue


from pyfyre.widgets.widget import Widget

class Button(Widget):
    """Creates a Button widget.

    Attributes
    ----------
    textContent : str (positional)
        The text content of the button.
    onClick : method
        The method you provided where when the button has been clicked
        or activate, the [onClick] method will be fired automatically.
    """
    
    def __init__(self, textContent, onClick=lambda: print(""), className="", props: dict=None):
        super().__init__("button", className=className, props=props)
        self.textContent = textContent
        self.onClick = onClick
    
    def dom(self):
        element = super().dom()
        element.textContent = self.textContent

        element.bind("click", self.onClick)
        
        return element



from pyfyre.widgets import *
from pyfyre.pyfyre import runApp
from pyfyre.router import Router
from src.main import HomePage
from src.components.header import Header
from src.playground import PlaygroundPage

class App(UsesState):
    def build(self):
        return Router(
            routes={
                "/": Home(),
                "/playground": Playground()
            }
        )
        
class Home(Container):
    def __init__(self):
        super().__init__(
            className = "home",
            children = [
                Header(),
                HomePage()
            ]
        )

class Playground(Container):
    def __init__(self):
        super().__init__(
            className = "home",
            children = [
                Header(),
                PlaygroundPage()
            ]
        )

runApp(
    App(),
    mount="app-mount"
)


from src.components.counterapp import CounterApp
from pyfyre.widgets import *
from src.components.experiment import Experiment

# You can make components by inherting a Widget class
# and initialize it on the `super().__init__()`.
class MyWebpage(UsesState):
    def __init__(self, greet):
        self.greet = greet

    def build(self):
        return Container(
            className = "container",
            children = [
                Text(
                    f"{self.greet} to PyFyre!",
                    className="title"
                ),
                Experiment(),
                CounterApp()
            ]
        )


from pyfyre.widgets.widget import Widget

from pyfyre.router import Router

class Link(Widget):
    """Creates a Link widget.

    This widget is used for Routers, if you want to redirect to a [Router] route,
    you must use this widget.

    Attributes
    ----------
    textContent : str (positional)
        The text content of the link widget.
    to : str (positional)
        The link of the route on the Router routes dictionary.
    """
    
    def __init__(self, textContent: str, to='/', external=False, className="", props: dict=None):
        super().__init__("a", className=className, props=props)
        self.textContent = textContent
        self.to = to
        self.external = external
    
    def dom(self):
        element = super().dom()
        element.textContent = self.textContent
        element.href = "#" if not self.external else self.to
        
        if not self.external:
            element.bind("click", self.navigate)
        
        return element

    def navigate(self, e):
        e.preventDefault()
        Router.push(self.to)


from pyfyre.globals import Globals
from pyfyre.globals.events import Events
from pyfyre.pyfyre import runApp

from browser import window, bind

class Router:
    """Router v0.3-alpha.
    
    Creates a wrapper object with window location
    history listener that when it's changed,
    the whole app rerenders and get the provided
    routes and map it correctly.

    Parameters
    ----------
    routes : dict
        A dictionary of object routes. Every value
        must inherit the [UsesState] object.

        For instance:
            Router(
                routes={
                    `"/": Home()`,
                    `"/about": About()`,

                    # Dynamic routing, `:<param>` tells the Router that the `about` path has `<param>` dynamic route.
                    `"/about/:id": AboutPerson()`
                }
            )

    Attributes
    ----------
    query : str
        If dynamic routing is allowed to the route, you can
        get the query of the route.

        For instance:
        `https://www.nicebookies.com/users/hi_mom`

        -----

        from pyfyre.router import Router

        Router(
            routes={
                `"/users/:slug": UserProfile()`,
            }
        )

        -----
        
        from pyfyre.router import Router

        class UserProfile(UsesState):
            def build(self):
                query = Router.query()

                return Text(f"Username: {query['slug']}")

    """

    def __init__(self, routes):
        self.routes: dict = routes

        self.initializeWindowLocation()
        self.listenRouteEvents()
        self.subscribeRoutes()
        
    def dom(self):

        if Globals.__LOC__ in self.routes.keys():
            return self.routes[Globals.__LOC__].dom()

        _, pathurl = self.get_params()

        for route in Globals.__ROUTES__:
            if route[0] == pathurl:
                return route[2].dom()

        # TODO: Create a 404.py that allows developers to customize 404 UI.
        # that can also be modified on the settings so they can
        # customize the 404.py name for their like.
        raise Exception("Path 404: Cannot find the path.")

    @staticmethod
    def query():
        query, pathurl = Router.get_params(Router)
        
        queryNames = {}

        for route in Globals.__ROUTES__:
            if route[0] == pathurl:
                queryNames[route[1]] = query

        return queryNames

    @staticmethod
    def push(location):
        Globals.__LOC__ = location

        Events.broadcast("change_route")
        window.history.pushState(None, None, location)

        @bind(window, 'popstate')
        def popState(e):
            Globals.__LOC__ = window.location.pathname
            Events.broadcast("change_route")
            e.preventDefault()

    def get_params(self):
        path = f"{Globals.__LOC__}"
        pathname = path.split('/')

        query = pathname[-1]
        pathurl = pathname[-2]

        return query, pathurl

    def initializeWindowLocation(self):
        if not Globals.__LOC__:
            Globals.__LOC__ = window.location.pathname

    def listenRouteEvents(self):
        if not "change_route" in Globals.__EVENTS__:
            Events.add("change_route")

            def changeRoute():
                runApp(Globals.__PARENT__)

            Events.addListener("change_route", changeRoute)

    def subscribeRoutes(self):
        for routeName, view in self.routes.items():
            pathnames = routeName.split('/')

            for pathname in pathnames:
                if len(pathname) <= 0: return
                
                if pathname[0] == ":":
                    pathname.replace(":", "")
                    Globals.__ROUTES__.append([pathnames[-2], pathname, view])


from pyfyre.widgets import *

from browser import document, window

codeListen = False

def runApp(app):
    try:
        out = document.getElementById("output-id")
        out.innerHTML = ""
        out.appendChild(app.dom())
    except Exception as e:
        out.innerHTML = ""
        out.innerHTML = "Error"

class PlaygroundPage(UsesState):
    def __init__(self):
        self.output = None
        self.controller = TextInputController()

        window.CodeListen.listen(self.compile)

    def compile(self):
        textArea = window.CodeMirrorAPI.codeMirror.getValue()

        exec(textArea)

    def build(self):

        return Container(
            className="flex flex-row w-full h-screen",
            children=[
                Container(
                    className="flex flex-col w-6/12",
                    children=[
                        TextInput(
                            className="",
                            controller=self.controller,
                            props={
                                "id": "code-editor",
                                "cols": 40,
                                "rows": 5 
                            },
                            multiline=True
                        ),
                    ]
                ),
                Container(
                    className="border-l-4",
                    children=[
                        Text("Result:", className="m-2 font-black"),
                        Container(
                            className="m-2",
                            props={"id": "output-id"},
                            children=[]
                        )
                    ]
                )
            ]
        )


from browser import ajax

class Ajax:
    """Access Brython builtin Ajax.

    ...

    Methods
    -------
    get(url, mode="json", then=None)
        Request a GET method to the url provided after the
        request, the [then] method will be fired.

    GET Parameters
    --------------
    url : str
        The URL where to request the GET
    mode : str
        Mode of the request
    then : method
        After the request, the [then] method provided
        will be fired passing the request object as the
        argument.
    """

    @staticmethod
    def get(url, mode="json", then=None):
        """Request a GET method to the url provided after the
        request, the [then] method will be fired.

        Parameters
        ----------
        url : str
            The URL where to request the GET
        mode : str
            Mode of the request
        then : method
            After the request, the [then] method provided
            will be fired passing the request object as the
            argument.
        
        """
        ajax.get(url, mode=mode, oncomplete=then)


from pyfyre.widgets import *

class Experiment(Container):
    def __init__(self):
        super().__init__(
            className = "test",
            children = [
                Text(
                    "Try to experiment, edit the `src/main.py` and wait for it to reload automatically on a liveserver. Anyways, I'm a component!",
                    className="desc"
                )
            ]
        )


from pyfyre.globals import Globals

class Events:
    """
    Creates an event and add listeners to it. Listeners are unfired methods.
    If an event fires off on `broadcastEvent` method, every listener of
    that event will be fired off.

    Methods
    -------
    add(event: str)
        Add a global event.

    addListener(event: str, listener: method)
        Add listener to the given [event].

    broadcast(event: str)
        Cast the event and fire all the listener methods

    """

    @staticmethod
    def add(event):
        """Create a global event.

        Parameters
        ----------
        event : str
            The name of the event.
        """
        Globals.__EVENTS__[event] = []

    @staticmethod
    def addListener(event, listener):
        """Add listener to a global event.

        Parameters
        ----------
        event : str
            The name of the event.
        listener : method
            The listener to the event that fires off when
            the event has been broadcast.
        """
        Globals.__EVENTS__[event].append(listener)

    @staticmethod
    def broadcast(event):
        """Broadcast the global event.

        Parameters
        ----------
        event : str
            The name of the event to be broadcast.
        """
        for func in Globals.__EVENTS__[event]:
            func()


from pyfyre.widgets.button import Button
from pyfyre.widgets.container import Container
from pyfyre.widgets.image import Image
from pyfyre.widgets.link import Link
from pyfyre.widgets.listbuilder import ListBuilder
from pyfyre.widgets.text import Text
from pyfyre.widgets.textinput import TextInput, TextInputController
from pyfyre.widgets.clickable import Clickable

from pyfyre.widgets.states import UsesState, State

__all__ = [
    'Button', 'Container', 'Image',
    'Link', 'ListBuilder', 'Clickable',
    'Text', 'TextInput', 'TextInputController',
    'Clickable', 'UsesState', 'State'
]


from pyfyre.widgets.widget import Widget

class Clickable(Widget):
    """Creates a Clickable widget.

    This wraps your widget into a clickable state. Just like InkWell
    in Flutter. Every widget (except Button) has no built-in onClick method
    so to allow then to be clickable, you must wrap them with this [Clickable] widget.

    Attributes
    ----------
    bind : Widget (positional)
        The widget you want to be clickable.
    onClick : method
        The method you provided where when the button has been clicked
        or activate, the [onClick] method will be fired automatically.
    """
    
    def __init__(self, bind, onClick, className="", props: dict=None):
        super().__init__("div", className=className, props=props)
        self.bind = bind
        self.onclick = onClick
    
    def dom(self):
        element = super().dom()
        
        element.appendChild(self.bind.dom())

        element.bind("click", self.onclick)
        
        return element


from pyfyre.widgets import *
from src.components.cta import CallToAction

# You can make components by inherting a Widget class
# and initialize it on the `super().__init__()`.
class HomePage(Container):
    def __init__(self):
        
        _features = [
            [
                "Developer friendly",
                "https://upload.wikimedia.org/wikipedia/commons/thumb/c/c3/Python-logo-notext.svg/1200px-Python-logo-notext.svg.png",
                "The language you already know.",
                "PyFyre allows you to develop apps with the language you already know, Python. Create reactive websites fast with ease. Build your apps in minutes, not days.",
                "Show me"
            ],
            [
                "Fast",
                "https://i.imgur.com/yI8tUab.png",
                "Fast transpilation time on the fly.",
                "PyFyre transpiles your code into small, vanilla JS that works for every browser with the power of Brython that stays fast, and always will be.",
                "Show me"
            ],
            [
                "Productive",
                "https://pbs.twimg.com/profile_images/909757546063323137/-RIWgodF_400x400.jpg",
                "Use Python Packages on the web.",
                "Use CPython's existing packages on PyPi.org/, develop sites quickly and efficiently with a large selection of packages.",
                "Visit pypi.org/"
            ],
        ]

        layouts = []

        for i in range(len(_features)):
            layout_1 = "md:flex-row-reverse" if i % 2 == 0 else "md:flex-row"
            layout_2 = "items-start" if i % 2 == 0 else "items-end"

            _layouts = []
            _layouts.append(layout_1)
            _layouts.append(layout_2)
            
            layouts.append(_layouts)

        def feature(i):
            return Container(
                className="",
                children=[
                    Container(
                        className=f"flex flex-col items-center px-5 py-16 mx-auto {layouts[i][0]} lg:px-28",
                        children=[
                            Container(
                                className="w-full mb-10 lg:w-1/3 lg:w-3/12 md:w-1/2",
                                children=[
                                    Image(
                                        className="object-cover object-center rounded-lg w-52 mx-auto md:w-auto",
                                        props={
                                            "loading": "lazy",
                                            "width": 500,
                                            "height": 500
                                        },
                                        src=_features[i][1]
                                    )
                                ]
                            ),
                            Container(
                                className=f"flex flex-col {layouts[i][1]} mb-16 text-left lg:flex-grow md:w-1/2 lg:pr-24 md:pr-16 md:mb-0",
                                children=[
                                    Text(
                                        className="mb-8 text-lg font-semibold tracking-widest text-black uppercase title-font text-[#333333]",
                                        textContent=_features[i][0]
                                    ),
                                    Text(
                                        className="mb-8 text-2xl font-black tracking-tighter text-black md:text-5xl title-font text-[#333333]",
                                        textContent=_features[i][2]
                                    ),
                                    Text(
                                        className="mb-8 text-base leading-relaxed text-left text-blueGray-600 text-[#333333]",
                                        textContent=_features[i][3]
                                    )
                                ]
                            )
                        ]
                    ),
                ]
            )

        super().__init__(
            className = "flex flex-col h-screen w-full",
            children = [
                Container(
                    className="flex flex-col w-full h-5/6",
                    children=[
                        Container(
                            className="flex flex-col w-full mt-28",
                            children=[
                                Container(
                                    className="flex flex-col",
                                    children=[
                                        Text(
                                            className = "text-5xl mx-auto font-bold bg-clip-text h-20 text-[#222222] xl:text-7xl",
                                            textContent = "Build reactive"
                                        ),
                                        Text(
                                            className = "text-5xl mx-auto font-bold bg-clip-text h-20 text-transparent bg-gradient-to-r from-cyan-500 to-blue-500 xl:text-7xl",
                                            textContent = "UI using Python."
                                        ),
                                    ]
                                ),
                                Text(
                                    className = "text-lg px-10 xl:text-xl mx-auto mt-5 xl:px-52 text-center text-[#3c3c3c]",
                                    textContent = "A fast, declarative, and incrementally adoptable Python web frontend framework for building reactive web user interfaces."
                                )
                            ]
                        ),
                        Container(
                            className="flex space-x-3 mx-auto mt-20 mb-44",
                            children=[
                                Link(
                                    className="bg-[#fab327] w-fit px-5 py-2 text-base rounded-3xl text-white cursor-pointer",
                                    textContent="Get started",
                                    to="https://pyfyre.gitbook.io/docs/",
                                    external=True
                                ),
                                Link(
                                    className="bg-[#f1f1f1] w-fit px-5 py-2 text-base rounded-xl text-[#474747] hover:text-[#333333] cursor-pointer",
                                    textContent="Playground",
                                    to="/playground"
                                )
                            ]
                        )
                    ]
                ),
                ListBuilder(
                    count=len(_features),
                    builder=feature
                ),
                CallToAction("Brython", "PyFyre is powered by Brython, a Python 3 implementation for client-side web programming.", "https://brython.info"),
            ]
        )


from pyfyre.widgets import *
from pyfyre.pyfyre import runApp
from src.main import MyWebpage

# This class is called at `runApp` method
# at the bottom. A class that inherits
# UsesState class has a `build` method that
# should return a Widget.
class App(UsesState):
    def __init__(self):
        # If you want to store a variable, you should store it
        # in `self` at the `__init__` method.
        self.greet = "Welcome"

    def build(self):

        # You can do some Pythonic things inside
        # the build method.

        return MyWebpage(self.greet)

# This is the main entry point of your app.
# Any class that inherits the UsesState class
# should be passed as the first parameter.
# The second parameter is the `mount`, it's
# an provided element id where your app will render.
# Try to take a look at `index.html` and you'll see
# a Div element with an id of `app-mount`.
runApp(
    App(),
    mount="app-mount"
)